"""`PrincipleMLChecker`."""

import copy
import logging
import os
import pickle
from collections import Counter
from collections.abc import Iterable
from typing import Any, Literal, Optional, Union, overload

import idstools.rule
import xgboost
from pandas import DataFrame, Series
from sklearn.metrics import f1_score, make_scorer, precision_score, recall_score
from sklearn.model_selection import (
    GridSearchCV,
    RepeatedStratifiedKFold,
    cross_val_score,
)
from sklearn.pipeline import Pipeline

from suricata_check._version import SURICATA_CHECK_DIR
from suricata_check.checkers.interface.checker import CheckerInterface
from suricata_check.checkers.principle._utils import get_message
from suricata_check.utils.checker import get_rule_option, get_rule_suboptions
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue

_PICKLE_PATH = os.path.join(SURICATA_CHECK_DIR, "data", "principle_ml_checker.pkl")
N_JOBS = 8


_logger = logging.getLogger(__name__)


COUNT_COLUMNS = (
    "flowbits.isset.count",
    "flowbits.isntoset.count",
    "flowint.isset.count",
    "flowint.isntoset.count",
    "xbits.isset.count",
    "xbits.uisnotset.count",
    "http.uri.count",
    "http.method.count",
    "dns.query.count",
    "content.count",
    "pcre.count",
    "startswith.count",
    "bsize.count",
    "depth.count",
    "urilen.count",
    "flow.from_server.count",
    "flow.to_server.count",
    "flow.from_client.count",
    "flow.to_client.count",
)
STRING_COLUMNS = ()
DROPDOWN_COLUMNS = (
    "proto",
    "threshold.type",
)
NUMERICAL_COLUMNS = ("threshold.count",)
SPLITTABLE_FEATURES = (
    "metadata",
    "flow",
    "threshold",
)
MSG_KEYWORDS = ("Suspicious", "CVE", "Vulnerability", "Response")
MSG_COLUMNS = ("msg.contains." + keyword for keyword in MSG_KEYWORDS)
IP_KEYWORDS = ("$HOME_NET", "$HTTP_SERVERS", "$EXTERNAL_NET", "any")
IP_COLUMNS = tuple(
    ["source_addr.contains." + keyword for keyword in IP_KEYWORDS]
    + ["dest_addr.contains." + keyword for keyword in IP_KEYWORDS]
)


PIPELINE = Pipeline(
    [
        (
            "classify",
            xgboost.XGBClassifier(),
        )
    ]
)
# https://shengyg.github.io/repository/machine%20learning/2017/02/25/Complete-Guide-to-Parameter-Tuning-xgboost.html
PARAM_GRID: list[dict] = [
    {
        # Fixed parameters for problem / desired complexity
        "classify__n_estimators": [1000],
        "classify__objective": ["binary:logistic"],
        ###
        # Parameters to optimize
        ## Learning rate
        "classify__eta": [0.01, 0.1, 0.3],
        ## Tree parameters
        "classify__subsample": [1.0],
        "classify__colsample_bytree": [0.25, 0.5, 0.75, 1.0],
        "classify__scale_pos_weight": [0.1, 0.25, 0.5, 1.0, 2.0, 4.0, 10.0],
        "classify__max_depth": [1, 3],
        "classify__min_child_weight": [1],
        "classify__gamma": [0, 0.1],
        ## Regularization
        "classify__lambda": [0, 0.01, 0.1],
        "classify__alpha": [0, 0.01, 0.1],
    },
]

PRECISION_WEIGHT = 10
SCORER = make_scorer(
    lambda y, y_pred: (PRECISION_WEIGHT + 1)
    / (
        PRECISION_WEIGHT / (precision_score(y, y_pred, zero_division=1) + 1e-10)  # type: ignore reportArgumentType
        + 1 / (recall_score(y, y_pred, zero_division=0) + 1e-10)  # type: ignore reportArgumentType
    )
)
SPLITTER = RepeatedStratifiedKFold(n_splits=2, n_repeats=10)
GRIDSEARCHCV = GridSearchCV(
    PIPELINE, PARAM_GRID, cv=SPLITTER, scoring=SCORER, error_score="raise", n_jobs=N_JOBS, verbose=1  # type: ignore reportArgumentType
)


class PrincipleMLChecker(CheckerInterface):
    """The `PrincipleChecker` contains several checks based on the Ruling the Unruly paper and target specificity and coverage.

    Codes Q000-Q009 report on non-adherence to rule design principles similar to Q000-Q009.
    Differently, they are the result of machine learning analysis of the rules.
    """

    count_columns = COUNT_COLUMNS
    string_columns = STRING_COLUMNS
    dropdown_columns = DROPDOWN_COLUMNS
    numerical_columns = NUMERICAL_COLUMNS
    splittable_features = SPLITTABLE_FEATURES
    msg_keywords = MSG_KEYWORDS
    msg_columns = MSG_COLUMNS
    ip_keywords = IP_KEYWORDS
    ip_columns = IP_COLUMNS

    codes = (
        "Q000",
        "Q001",
        "Q002",
        "Q003",
        "Q004",
        "Q005",
    )

    enabled_by_default = (
        False  # Since the checker is relatively slow, it is disabled by default
    )

    _dtypes: Optional[dict[str, Any]] = None
    _models: dict[str, Pipeline] = {}

    def __new__(
        cls: type["PrincipleMLChecker"],
        filepath: Optional[str] = _PICKLE_PATH,
        *args: tuple,
        **kwargs: dict,
    ) -> "PrincipleMLChecker":
        """Returns a new or unpickled instance of the class."""
        if filepath:
            if os.path.exists(filepath):
                with open(filepath, "rb") as f:
                    inst = pickle.load(f)

                # BEGIN LEGACY CODE
                # CAN BE REMOVED AFTER TRAINING NEW PKL
                if hasattr(inst, "models"):
                    inst._models = inst.models  # noqa: SLF001
                    inst._dtypes = inst.dtypes  # noqa: SLF001
                # END LEGACY CODE

                if not inst.__class__.__name__ == cls.__name__:
                    _logger.error("Unpickled object is not of type %s", cls)
                    inst = super().__new__(cls, *args, **kwargs)  # type: ignore reportargumentType
                elif (
                    not hasattr(inst, "_models")
                    or len(
                        inst._models  # noqa: SLF001 type: ignore reportAttributeAccess
                    )
                    == 0
                ):
                    _logger.error("Unpickled object does not have trained models")
                    inst = super().__new__(cls, *args, **kwargs)  # type: ignore reportargumentType
                else:
                    if "include" in kwargs:
                        inst.include = kwargs["include"]
                    _logger.info("Unpickled object with trained models successfully")
            else:
                _logger.warning("No model found for PrincipleMLChecker at %s", filepath)
                inst = super().__new__(cls, *args, **kwargs)  # type: ignore reportargumentType
        else:
            inst = super().__new__(cls, *args, **kwargs)  # type: ignore reportargumentType

        return inst

    def __getnewargs__(
        self: "PrincipleMLChecker",
    ) -> tuple:
        """Returns the arguments to be passed to the __new__ method when unpickling."""
        return (None,)

    def _check_rule(
        self: "PrincipleMLChecker",
        rule: idstools.rule.Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if len(self._models) == 0:
            return issues

        for code, model in self._models.items():
            if model.predict(self._get_features(rule, True))[0]:
                issues.append(
                    Issue(
                        code=code,
                        message=get_message(code),
                    )
                )

        return issues

    def train(  # noqa: C901
        self: "PrincipleMLChecker",
        df: DataFrame,
        rule_col: str = "rule.rule",
        principle_cols: dict[str, str] = {
            "Q000": "labelled.no_proxy",
            "Q001": "labelled.success",
            "Q002": "labelled.thresholded",
            "Q003": "labelled.exceptions",
            "Q004": "labelled.generalized_match_content",
            "Q005": "labelled.generalized_match_location",
        },
        reuse_models: bool = False,
    ) -> None:
        """Train several models for the checker to detect issues in rules.

        The checker class with trained models is stored in a pickle file (`_PICKLE_PATH`).
        """
        self._dtypes = None
        if not reuse_models:
            self._models = {}

        # Extract features and determine feature dtypes
        X_train = self._get_train_df(df[rule_col])  # noqa: N806

        for col in X_train.columns:
            try:
                X_train[col].var()
                _logger.debug("Detected column: %s", col)
            except:
                _logger.error("Error with column %s", col)
                _logger.error(X_train[col])
                raise

        # # Drop zero variance columns
        X_train = X_train.drop(  # noqa: N806
            X_train.columns[(X_train.fillna(-1337).var(axis=0) <= 0)].to_list(),  # type: ignore reportAttributeAccessIssue
            axis=1,
        )

        # Drop columns with too few occurrences of possible values
        for col in X_train.columns:
            if (
                not col.endswith(".count")
                and not col.endswith(".num")
                and not col.endswith(".len")
            ):
                if X_train[col].value_counts().min() <= 1:
                    X_train = X_train.drop(  # noqa: N806
                        [col],
                        axis=1,
                    )

        for col in X_train.columns:
            try:
                X_train[col].var()
                _logger.info("Using column: %s", col)
            except:
                _logger.error("Error with column %s", col)
                _logger.error(X_train[col])
                raise

        # Store used features and their dtypes
        self._dtypes = X_train.dtypes.to_dict()
        _logger.debug(self._dtypes)

        # Redo feature extraction now that FE parameters are set
        X_train = self._get_train_df(df[rule_col])  # noqa: N806

        _logger.info(
            "Training model with features: [%s]",
            ", ".join([str(x) for x in X_train.columns]),
        )

        _logger.info(X_train)

        for code, col in principle_cols.items():
            y_true = df[col].to_numpy() == 0

            if not reuse_models or code not in self._models:
                # Train new model with grid search to find optimal parameters
                gridsearchcv: GridSearchCV = copy.deepcopy(GRIDSEARCHCV)

                gridsearchcv.fit(X_train, y_true)

                _logger.info("Code %s params: %s", code, gridsearchcv.best_params_)
                _logger.info(
                    "Code %s Weighted F1-score: %s", code, gridsearchcv.best_score_
                )

                self._models[code] = gridsearchcv.best_estimator_

            precision = cross_val_score(
                self._models[code],
                X_train,
                y_true,
                scoring=make_scorer(precision_score, zero_division=0.0),
                cv=SPLITTER,
                n_jobs=N_JOBS,
            ).mean()
            recall = cross_val_score(
                self._models[code],
                X_train,
                y_true,
                scoring=make_scorer(recall_score, zero_division=0.0),
                cv=SPLITTER,
                n_jobs=N_JOBS,
            ).mean()
            f1 = cross_val_score(
                self._models[code],
                X_train,
                y_true,
                scoring=make_scorer(f1_score, zero_division=0.0),
                cv=SPLITTER,
                n_jobs=N_JOBS,
            ).mean()
            _logger.info("Code %s Precision score: %s", code, precision)
            _logger.info("Code %s Recall score: %s", code, recall)
            _logger.info("Code %s F1-score: %s", code, f1)

            # Refit model with training data.
            self._models[code].fit(X_train, y_true)

        pickle.dump(self, open(_PICKLE_PATH, "wb"))

    def _get_train_df(self: "PrincipleMLChecker", rules: Iterable[str]) -> DataFrame:
        feature_vectors = []
        for rule in rules:
            parsed_rule = idstools.rule.parse(rule)
            feature_vectors.append(self._get_features(parsed_rule, False))

        return DataFrame(feature_vectors)

    def _get_raw_features(  # noqa: C901
        self: "PrincipleMLChecker", rule: idstools.rule.Rule
    ) -> Series:
        d: dict[str, Optional[Union[str, int]]] = {
            "proto": get_rule_option(rule, "proto")
        }

        options = rule["options"]

        for option in options:
            d[option["name"]] = option["value"]

        counter = Counter([option["name"] for option in options])
        for option, count in counter.items():
            d[option + ".count"] = count

        for option in options:
            if option["name"] not in self.splittable_features:
                continue

            suboptions = [
                {"name": k, "value": v}
                for k, v in get_rule_suboptions(rule, option["name"], warn=False)
            ]

            if len(suboptions) == 0:
                continue

            for suboption in suboptions:
                d[option["name"] + "." + suboption["name"]] = suboption["value"]

            counter = Counter([suboption["name"] for suboption in suboptions])
            for suboption, count in counter.items():
                d[option["name"] + "." + suboption + ".count"] = count

        msg = get_rule_option(rule, "msg")
        assert msg is not None
        msg = msg.lower()
        for col, keyword in zip(self.msg_columns, self.msg_keywords):
            d[col] = keyword.lower() in msg

        source_addr = get_rule_option(rule, "source_addr")
        assert source_addr is not None
        source_addr = source_addr.lower()
        for keyword in self.ip_keywords:
            col = "source_addr.contains." + keyword
            d[col] = keyword.lower() in source_addr

        dest_addr = get_rule_option(rule, "dest_addr")
        assert dest_addr is not None
        dest_addr = dest_addr.lower()
        for keyword in self.ip_keywords:
            col = "dest_addr.contains." + keyword
            d[col] = keyword.lower() in dest_addr

        return Series(d)

    def _preprocess_features(self: "PrincipleMLChecker", data: Series) -> Series:
        original_cols: set[str] = set(data.index)

        for col in self.string_columns:
            if col not in data:
                continue
            data[col + ".len"] = len(data[col])
            data = data.drop(col)

        for col in self.dropdown_columns:
            if col not in data:
                continue
            data[col + "." + data[col] + ".bool"] = 1
            data = data.drop(col)

        for col in self.numerical_columns:
            if col not in data:
                continue
            data[col + ".num"] = float(data[col])  # type: ignore reportArgumentType
            data = data.drop(col)

        remaining_cols = (
            original_cols
            - set(self.count_columns)
            - set(self.string_columns)
            - set(self.dropdown_columns)
            - set(self.numerical_columns)
            - set(self.msg_columns)
            - set(self.ip_columns)
        )

        for col in remaining_cols:
            data = data.drop(col)

        return data

    @overload
    def _get_features(
        self: "PrincipleMLChecker", rule: idstools.rule.Rule, frame: Literal[True]
    ) -> DataFrame:
        pass

    @overload
    def _get_features(
        self: "PrincipleMLChecker", rule: idstools.rule.Rule, frame: Literal[False]
    ) -> Series:
        pass

    def _get_features_frame(self: "PrincipleMLChecker", features: Series) -> DataFrame:
        features_frame = features.to_frame().transpose()

        if self._dtypes is None:
            return features_frame

        for col, dtype in self._dtypes.items():
            if features_frame.dtypes[col] != dtype:
                features_frame[col] = features_frame[col].astype(dtype)

        return features_frame

    def _get_features(
        self: "PrincipleMLChecker", rule: idstools.rule.Rule, frame: bool
    ) -> Union[Series, DataFrame]:
        features: Series = self._get_raw_features(rule)
        features = self._preprocess_features(features)

        features["custom.negated.count"] = rule["raw"].count(':!"')

        if self._dtypes is None:
            return features

        for col, dtype in self._dtypes.items():
            if col not in features:
                if col.endswith(".count"):
                    features[col] = 0
                elif col.endswith(".bool"):
                    features[col] = 0
                elif col.endswith(".num"):
                    features[col] = -1
                else:
                    _logger.error(
                        "Unsure how to handle missing feature %s of type %s",
                        col,
                        dtype,
                    )

        features = features[list(self._dtypes.keys())]  # type: ignore reportAssignmentType

        if not frame:
            return features

        return self._get_features_frame(features)



"""The `suricata_check.suricata_check` module contains the command line utility and the main program logic."""

import atexit
import io
import logging
import logging.handlers
import multiprocessing
import os
import pkgutil
import sys
from collections import defaultdict
from collections.abc import Mapping, Sequence
from functools import lru_cache
from typing import (
    Literal,
    Optional,
)

import click
import idstools.rule
import tabulate

# Add suricata-check to the front of the PATH, such that the version corresponding to the CLI is used.
_suricata_check_path = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
if sys.path[0] != _suricata_check_path:
    sys.path.insert(0, _suricata_check_path)

from suricata_check import __version__, get_dependency_versions  # noqa: E402
from suricata_check.checkers.interface import CheckerInterface  # noqa: E402
from suricata_check.checkers.interface.dummy import DummyChecker  # noqa: E402
from suricata_check.utils._click import ClickHandler  # noqa: E402
from suricata_check.utils._path import find_rules_file  # noqa: E402
from suricata_check.utils.checker import check_rule_option_recognition  # noqa: E402
from suricata_check.utils.checker_typing import (  # noqa: E402
    EXTENSIVE_SUMMARY_TYPE,
    ISSUES_TYPE,
    RULE_REPORTS_TYPE,
    RULE_SUMMARY_TYPE,
    SIMPLE_SUMMARY_TYPE,
    InvalidRuleError,
    OutputReport,
    OutputSummary,
    RuleReport,
    get_all_subclasses,
)
from suricata_check.utils.regex import get_regex_provider, is_valid_rule  # noqa: E402

LOG_LEVELS = ("DEBUG", "INFO", "WARNING", "ERROR")
LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR"]

_logger = logging.getLogger(__name__)

_regex_provider = get_regex_provider()

# Global variable to check if extensions have already been imported in case get_checkers() is called multiple times.
suricata_check_extensions_imported = False


@click.command()
@click.option(
    "--out",
    "-o",
    default=".",
    help="Path to suricata-check output folder.",
    show_default=True,
)
@click.option(
    "--rules",
    "-r",
    default=".",
    help="Path to Suricata rules to provide check on.",
    show_default=True,
)
@click.option(
    "--single-rule",
    "-s",
    help="A single Suricata rule to be checked",
    show_default=False,
)
@click.option(
    "--log-level",
    default="INFO",
    help=f"Verbosity level for logging. Can be one of {LOG_LEVELS}",
    show_default=True,
)
@click.option(
    "--evaluate-disabled",
    default=False,
    help="Flag to evaluate disabled rules.",
    show_default=True,
    is_flag=True,
)
@click.option(
    "--include-all",
    "-a",
    default=False,
    help="Flag to indicate all checker codes should be enabled.",
    show_default=True,
    is_flag=True,
)
@click.option(
    "--include",
    "-i",
    default=(),
    help="List of all checker codes to enable.",
    show_default=True,
    multiple=True,
)
@click.option(
    "--exclude",
    "-e",
    default=(),
    help="List of all checker codes to disable.",
    show_default=True,
    multiple=True,
)
def main(  # noqa: PLR0913
    out: str = ".",
    rules: str = ".",
    single_rule: Optional[str] = None,
    log_level: LogLevel = "DEBUG",
    evaluate_disabled: bool = False,
    include_all: bool = False,
    include: tuple[str, ...] = (),
    exclude: tuple[str, ...] = (),
) -> None:
    """The `suricata-check` command processes all rules inside a rules file and outputs a list of detected issues.

    Raises:
      BadParameter: If provided arguments are invalid.

      RuntimeError: If no checkers could be automatically discovered.

    """
    # Verify that out argument is valid
    if os.path.exists(out) and not os.path.isdir(out):
        raise click.BadParameter(f"Error: {out} is not a directory.")

    # Verify that log_level argument is valid
    if log_level not in LOG_LEVELS:
        raise click.BadParameter(f"Error: {log_level} is not a valid log level.")

    # Create out directory if non-existent
    if not os.path.exists(out):
        os.makedirs(out)

    # Setup logging from a seperate thread
    queue = multiprocessing.Manager().Queue()
    queue_handler = logging.handlers.QueueHandler(queue)

    click_handler = ClickHandler()
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        handlers=(queue_handler, click_handler),
        force=os.environ.get("SURICATA_CHECK_FORCE_LOGGING", False) == "TRUE",
    )

    file_handler = logging.FileHandler(
        filename=os.path.join(out, "suricata-check.log"),
        delay=True,
    )
    queue_listener = logging.handlers.QueueListener(
        queue,
        file_handler,
        respect_handler_level=True,
    )

    def _at_exit() -> None:
        """Cleans up logging listener and handlers before exiting."""
        queue_listener.enqueue_sentinel()
        queue_listener.stop()
        file_handler.flush()
        file_handler.close()
        atexit.unregister(_at_exit)

    atexit.register(_at_exit)

    queue_listener.start()

    # Log the arguments:
    _logger.info("Running suricata-check with the following arguments:")
    _logger.info("out: %s", out)
    _logger.info("rules: %s", rules)
    _logger.info("single_rule: %s", single_rule)
    _logger.info("log_level: %s", log_level)
    _logger.info("evaluate_disabled: %s", evaluate_disabled)
    _logger.info("include_all: %s", include_all)
    _logger.info("include: %s", include)
    _logger.info("exclude: %s", exclude)

    # Log the environment:
    _logger.debug("Platform: %s", sys.platform)
    _logger.debug("Python version: %s", sys.version)
    _logger.debug("suricata-check path: %s", _suricata_check_path)
    _logger.debug("suricata-check version: %s", __version__)
    for package, version in get_dependency_versions().items():
        _logger.debug("Dependency %s version: %s", package, version)

    # Verify that include and exclude arguments are valid
    if include_all and len(include) > 0:
        raise click.BadParameter(
            "Error: Cannot use --include-all and --include together."
        )
    if include_all:
        include = (".*",)

    checkers = get_checkers(include, exclude)

    if single_rule is not None:
        __main_single_rule(out, single_rule, checkers)

        # Return here so no rules file is processed.
        _at_exit()
        return

    # Check if the rules argument is valid and find the rules file
    rules = find_rules_file(rules)

    output = process_rules_file(rules, evaluate_disabled, checkers=checkers)

    __write_output(output, out)

    _at_exit()


def __main_single_rule(
    out: str, single_rule: str, checkers: Optional[Sequence[CheckerInterface]]
) -> None:
    rule: Optional[idstools.rule.Rule] = idstools.rule.parse(single_rule)

    # Verify that a rule was parsed correctly.
    if rule is None:
        msg = f"Error parsing rule from user input: {single_rule}"
        _logger.critical(msg)
        raise click.BadParameter(f"Error: {msg}")

    if not is_valid_rule(rule):
        msg = f"Error parsing rule from user input: {single_rule}"
        _logger.critical(msg)
        raise click.BadParameter(f"Error: {msg}")

    _logger.debug("Processing rule: %s", rule["sid"])

    rule_report = analyze_rule(rule, checkers=checkers)

    __write_output(OutputReport(rules=[rule_report]), out)


def __write_output(
    output: OutputReport,
    out: str,
) -> None:
    _logger.info(
        "Writing output to suricata-check.jsonl and suricata-check-fast.log in %s",
        os.path.abspath(out),
    )
    with (
        open(
            os.path.join(out, "suricata-check.jsonl"),
            "w",
            buffering=io.DEFAULT_BUFFER_SIZE,
        ) as jsonl_fh,
        open(
            os.path.join(out, "suricata-check-fast.log"),
            "w",
            buffering=io.DEFAULT_BUFFER_SIZE,
        ) as fast_fh,
    ):
        rules: RULE_REPORTS_TYPE = output.rules
        jsonl_fh.write("\n".join([str(rule) for rule in rules]))

        for rule_report in rules:
            rule: idstools.rule.Rule = rule_report.rule
            line: Optional[int] = rule_report.line
            issues: ISSUES_TYPE = rule_report.issues
            for issue in issues:
                code = issue.code
                issue_msg = issue.message.replace("\n", " ")

                msg = f"[{code}] Line {line}, sid {rule['sid']}: {issue_msg}"
                fast_fh.write(msg + "\n")
                click.secho(msg, color=True, fg="blue")

    if output.summary is not None:
        with open(
            os.path.join(out, "suricata-check-stats.log"),
            "w",
            buffering=io.DEFAULT_BUFFER_SIZE,
        ) as stats_fh:
            summary: OutputSummary = output.summary

            overall_summary: SIMPLE_SUMMARY_TYPE = summary.overall_summary

            n_issues = overall_summary["Total Issues"]
            n_rules = (
                overall_summary["Rules with Issues"]
                + overall_summary["Rules without Issues"]
            )

            stats_fh.write(
                tabulate.tabulate(
                    (
                        (
                            k,
                            v,
                            (
                                "{:.0%}".format(v / n_rules)
                                if k.startswith("Rules ") and n_rules > 0
                                else "-"
                            ),
                        )
                        for k, v in overall_summary.items()
                    ),
                    headers=(
                        "Count",
                        "Percentage of Rules",
                    ),
                )
                + "\n\n",
            )

            click.secho(
                f"Total issues found: {overall_summary['Total Issues']}",
                color=True,
                bold=True,
                fg="blue",
            )
            click.secho(
                f"Rules with Issues found: {overall_summary['Rules with Issues']}",
                color=True,
                bold=True,
                fg="blue",
            )

            issues_by_group: SIMPLE_SUMMARY_TYPE = summary.issues_by_group

            stats_fh.write(
                tabulate.tabulate(
                    (
                        (k, v, "{:.0%}".format(v / n_issues) if n_issues > 0 else "-")
                        for k, v in issues_by_group.items()
                    ),
                    headers=(
                        "Count",
                        "Percentage of Total Issues",
                    ),
                )
                + "\n\n",
            )

            issues_by_type: EXTENSIVE_SUMMARY_TYPE = summary.issues_by_type
            for checker, checker_issues_by_type in issues_by_type.items():
                stats_fh.write(" " + checker + " " + "\n")
                stats_fh.write("-" * (len(checker) + 2) + "\n")
                stats_fh.write(
                    tabulate.tabulate(
                        (
                            (
                                k,
                                v,
                                "{:.0%}".format(v / n_rules) if n_rules > 0 else "-",
                            )
                            for k, v in checker_issues_by_type.items()
                        ),
                        headers=(
                            "Count",
                            "Percentage of Rules",
                        ),
                    )
                    + "\n\n",
                )


def process_rules_file(
    rules: str,
    evaluate_disabled: bool,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> OutputReport:
    """Processes a rule file and returns a list of rules and their issues.

    Args:
    rules: A path to a Suricata rules file.
    evaluate_disabled: A flag indicating whether disabled rules should be evaluated.
    checkers: The checkers to be used when processing the rule file.

    Returns:
        A list of rules and their issues.

    Raises:
        RuntimeError: If no checkers could be automatically discovered.

    """
    if checkers is None:
        checkers = get_checkers()

    output = OutputReport()

    with (
        open(
            os.path.normpath(rules),
            buffering=io.DEFAULT_BUFFER_SIZE,
        ) as rules_fh,
    ):
        if len(checkers) == 0:
            msg = "No checkers provided for processing rules."
            _logger.error(msg)
            raise RuntimeError(msg)

        _logger.info("Processing rule file: %s", rules)

        for number, line in enumerate(rules_fh.readlines(), start=1):
            if line.startswith("#"):
                if evaluate_disabled:
                    # Verify that this line is a rule and not a comment
                    if idstools.rule.parse(line) is None:
                        # Log the comment since it may be a invalid rule
                        _logger.warning("Ignoring comment on line %i: %s", number, line)
                        continue
                else:
                    # Skip the rule
                    continue

            # Skip whitespace
            if len(line.strip()) == 0:
                continue

            rule: Optional[idstools.rule.Rule] = idstools.rule.parse(line)

            # Verify that a rule was parsed correctly.
            if rule is None:
                _logger.error("Error parsing rule on line %i: %s", number, line)
                continue

            if not is_valid_rule(rule):
                _logger.error("Invalid rule on line %i: %s", number, line)
                continue

            _logger.debug("Processing rule: %s on line %i", rule["sid"], number)

            rule_report: RuleReport = analyze_rule(
                rule,
                checkers=checkers,
            )
            rule_report.line = number
            output.rules.append(rule_report)

    _logger.info("Completed processing rule file: %s", rules)

    output.summary = __summarize_output(output, checkers)

    return output


def _import_extensions() -> None:
    global suricata_check_extensions_imported  # noqa: PLW0603
    if suricata_check_extensions_imported is True:
        return

    for module in pkgutil.iter_modules():
        if module.name.startswith("suricata_check_"):
            try:
                imported_module = __import__(module.name)
                _logger.info(
                    "Detected and successfully imported suricata-check extension %s with version %s.",
                    module.name.replace("_", "-"),
                    getattr(imported_module, "__version__"),
                )
            except ImportError:
                _logger.warning(
                    "Detected potential suricata-check extension %s but failed to import it.",
                    module.name.replace("_", "-"),
                )
    suricata_check_extensions_imported = True


@lru_cache(maxsize=1)
def get_checkers(
    include: Sequence[str] = (".*",),
    exclude: Sequence[str] = (),
) -> Sequence[CheckerInterface]:
    """Auto discovers all available checkers that implement the CheckerInterface.

    Returns:
    A list of available checkers that implement the CheckerInterface.

    """
    # Check for extensions and try to import them
    _import_extensions()

    checkers: list[CheckerInterface] = []
    for checker in get_all_subclasses(CheckerInterface):
        if checker.__name__ == DummyChecker.__name__:
            continue

        # Initialize DummyCheckers to retrieve error messages.
        if issubclass(checker, DummyChecker):
            checker()

        enabled, relevant_codes = __get_checker_enabled(checker, include, exclude)

        if enabled:
            checkers.append(checker(include=relevant_codes))

        else:
            _logger.info("Checker %s is disabled.", checker.__name__)

    _logger.info(
        "Discovered and enabled checkers: [%s]",
        ", ".join([c.__class__.__name__ for c in checkers]),
    )
    if len(checkers) == 0:
        _logger.warning(
            "No checkers were enabled. Check the include and exclude arguments."
        )

    # Perform a uniqueness check on the codes emmitted by the checkers
    for checker1 in checkers:
        for checker2 in checkers:
            if checker1 == checker2:
                continue
            if not set(checker1.codes).isdisjoint(checker2.codes):
                msg = f"Checker {checker1.__class__.__name__} and {checker2.__class__.__name__} have overlapping codes."
                _logger.error(msg)

    return sorted(checkers, key=lambda x: x.__class__.__name__)


def __get_checker_enabled(
    checker: type[CheckerInterface],
    include: Sequence[str],
    exclude: Sequence[str],
) -> tuple[bool, set[str]]:
    enabled = checker.enabled_by_default

    relevant_codes = set(checker.codes)

    if len(include) > 0:
        for regex in include:
            relevant_codes = set(
                filter(
                    lambda code: _regex_provider.compile("^" + regex + "$").match(code)
                    is not None,
                    relevant_codes,
                )
            )
            if len(relevant_codes) > 0:
                enabled = True
    for regex in exclude:
        relevant_codes = set(
            filter(
                lambda code: _regex_provider.compile("^" + regex + "$").match(code)
                is None,
                relevant_codes,
            )
        )

    if len(relevant_codes) == 0:
        enabled = False

    return enabled, relevant_codes


def analyze_rule(
    rule: idstools.rule.Rule,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> RuleReport:
    """Checks a rule and returns a dictionary containing the rule and a list of issues found.

    Args:
    rule: The rule to be checked.
    checkers: The checkers to be used to check the rule.

    Returns:
    A list of issues found in the rule.
    Each issue is typed as a `dict`.

    Raises:
    InvalidRuleError: If the rule does not follow the Suricata syntax.

    """
    if not is_valid_rule(rule):
        raise InvalidRuleError(rule["raw"])

    check_rule_option_recognition(rule)

    if checkers is None:
        checkers = get_checkers()

    rule_report: RuleReport = RuleReport(rule=rule)

    for checker in checkers:
        rule_report.add_issues(checker.check_rule(rule))

    rule_report.summary = __summarize_rule(rule_report, checkers)

    return rule_report


def __summarize_rule(
    rule: RuleReport,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> RULE_SUMMARY_TYPE:
    """Summarizes the issues found in a rule.

    Args:
    rule: The rule output dictionary to be summarized.
    checkers: The checkers to be used to check the rule.

    Returns:
    A dictionary containing a summary of all issues found in the rule.

    """
    if checkers is None:
        checkers = get_checkers()

    summary = {}

    issues: ISSUES_TYPE = rule.issues
    summary["total_issues"] = len(issues)
    summary["issues_by_group"] = defaultdict(int)
    for issue in issues:
        checker = issue.checker
        summary["issues_by_group"][checker] += 1

    # Ensure also checkers without issues are included in the report.
    for checker in checkers:
        if checker.__class__.__name__ not in summary["issues_by_group"]:
            summary["issues_by_group"][checker.__class__.__name__] = 0

    # Sort dictionaries for deterministic output
    summary["issues_by_group"] = __sort_mapping(summary["issues_by_group"])

    return summary


def __summarize_output(
    output: OutputReport,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> OutputSummary:
    """Summarizes the issues found in a rules file.

    Args:
    output: The unsammarized output of the rules file containing all rules and their issues.
    checkers: The checkers to be used to check the rule.

    Returns:
    A dictionary containing a summary of all issues found in the rules file.

    """
    if checkers is None:
        checkers = get_checkers()

    return OutputSummary(
        overall_summary=__get_overall_summary(output),
        issues_by_group=__get_issues_by_group(output, checkers),
        issues_by_type=__get_issues_by_type(output, checkers),
    )


def __get_overall_summary(
    output: OutputReport,
) -> SIMPLE_SUMMARY_TYPE:
    overall_summary = {
        "Total Issues": 0,
        "Rules with Issues": 0,
        "Rules without Issues": 0,
    }

    rules: RULE_REPORTS_TYPE = output.rules
    for rule in rules:
        issues: ISSUES_TYPE = rule.issues
        overall_summary["Total Issues"] += len(issues)

        if len(issues) == 0:
            overall_summary["Rules without Issues"] += 1
        else:
            overall_summary["Rules with Issues"] += 1

    return overall_summary


def __get_issues_by_group(
    output: OutputReport,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> SIMPLE_SUMMARY_TYPE:
    if checkers is None:
        checkers = get_checkers()

    issues_by_group = defaultdict(int)

    # Ensure also checkers and codes without issues are included in the report.
    for checker in checkers:
        issues_by_group[checker.__class__.__name__] = 0

    rules: RULE_REPORTS_TYPE = output.rules
    for rule in rules:
        issues: ISSUES_TYPE = rule.issues

        for issue in issues:
            checker = issue.checker
            if checker is not None:
                issues_by_group[checker] += 1

    return __sort_mapping(issues_by_group)


def __get_issues_by_type(
    output: OutputReport,
    checkers: Optional[Sequence[CheckerInterface]] = None,
) -> EXTENSIVE_SUMMARY_TYPE:
    if checkers is None:
        checkers = get_checkers()
    issues_by_type: EXTENSIVE_SUMMARY_TYPE = defaultdict(lambda: defaultdict(int))

    # Ensure also checkers and codes without issues are included in the report.
    for checker in checkers:
        for code in checker.codes:
            issues_by_type[checker.__class__.__name__][code] = 0

    rules: RULE_REPORTS_TYPE = output.rules
    for rule in rules:
        issues: ISSUES_TYPE = rule.issues

        checker_codes = defaultdict(lambda: defaultdict(int))
        for issue in issues:
            checker = issue.checker
            if checker is not None:
                code = issue.code
                checker_codes[checker][code] += 1

        for checker, codes in checker_codes.items():
            for code, count in codes.items():
                issues_by_type[checker][code] += count

    for key in issues_by_type:
        issues_by_type[key] = __sort_mapping(issues_by_type[key])

    return __sort_mapping(issues_by_type)


def __sort_mapping(mapping: Mapping) -> dict:
    return {key: mapping[key] for key in sorted(mapping.keys())}


if __name__ == "__main__":
    main()



"""The `suricata_check.checkers.principle` modules contains several checkers based on the Ruling the Unruly paper.

Reference: TODO
"""

from suricata_check.checkers.interface.dummy import DummyChecker
from suricata_check.checkers.principle.principle import PrincipleChecker

try:
    from suricata_check.checkers.principle.ml import PrincipleMLChecker  # type: ignore reportAssignmentType
except ImportError:

    class PrincipleMLChecker(DummyChecker):
        """Dummy class to prevent runtime errors on import."""


__all__ = [
    "PrincipleChecker",
    "PrincipleMLChecker",
]



import logging

import click


class ClickHandler(logging.Handler):
    """Handler to color and write logging messages for the click module."""

    def emit(self: "ClickHandler", record: logging.LogRecord) -> None:
        """Log the record via click stdout with appropriate colors."""
        msg = self.format(record)

        if logging.getLevelName(record.levelno) == "DEBUG":
            click.secho(msg, color=True, dim=True)
        if logging.getLevelName(record.levelno) == "INFO":
            click.secho(msg, color=True)
        if logging.getLevelName(record.levelno) == "WARNING":
            click.secho(msg, color=True, bold=True, fg="yellow")
        if logging.getLevelName(record.levelno) == "ERROR":
            click.secho(msg, color=True, bold=True, fg="red")
        if logging.getLevelName(record.levelno) == "CRITICAL":
            click.secho(msg, color=True, bold=True, blink=True, fg="red")



"""`PrincipleChecker`."""

from typing import Optional

import idstools.rule

from suricata_check.checkers.interface.checker import CheckerInterface
from suricata_check.checkers.principle._utils import get_message
from suricata_check.utils.checker import (
    count_rule_options,
    get_rule_option,
    get_rule_options,
    is_rule_option_equal_to,
    is_rule_option_equal_to_regex,
    is_rule_option_set,
    is_rule_suboption_set,
)
from suricata_check.utils.checker_typing import ISSUES_TYPE, Issue
from suricata_check.utils.regex import (
    ALL_DETECTION_KEYWORDS,
    BUFFER_KEYWORDS,
    CONTENT_KEYWORDS,
    IP_ADDRESS_REGEX,
    OTHER_PAYLOAD_KEYWORDS,
    SIZE_KEYWORDS,
    get_options_regex,
    get_regex_provider,
    get_rule_body,
)

_regex_provider = get_regex_provider()

_BITS_ISSET_REGEX = _regex_provider.compile(r"^\s*isset\s*,.*$")
_BITS_ISNOTSET_REGEX = _regex_provider.compile(r"^\s*isnotset\s*,.*$")
_FLOWINT_ISSET_REGEX = _regex_provider.compile(r"^.*,\s*isset\s*,.*$")
_FLOWINT_ISNOTSET_REGEX = _regex_provider.compile(r"^.*,\s*isnotset\s*,.*$")
_THRESHOLD_LIMITED_REGEX = _regex_provider.compile(r"^.*type\s+(limit|both).*$")
_FLOWBITS_ISNOTSET_REGEX = _regex_provider.compile(r"^\s*isnotset.*$")
_HTTP_URI_QUERY_PARAMETER_REGEX = _regex_provider.compile(
    rf"^\(.*\s+http\.uri\s*;\s*content\s*:\s*\"[^\"]*\?([^\"]|\\\")+\"\s*;((?!.*{get_options_regex(CONTENT_KEYWORDS).pattern}).*)|((?!.*{get_options_regex(CONTENT_KEYWORDS).pattern}).*\s+{get_options_regex(BUFFER_KEYWORDS).pattern}\s*;.*)\)$"
)
_PROXY_MSG_REGEX = _regex_provider.compile(
    r"^.*(Suspicious).*$", flags=_regex_provider.IGNORECASE
)
_SPECIFIC_MSG_REGEX = _regex_provider.compile(
    r"^.*(CVE|Vulnerability).*$", flags=_regex_provider.IGNORECASE
)


class PrincipleChecker(CheckerInterface):
    """The `PrincipleChecker` contains several checks based on the Ruling the Unruly paper and target specificity and coverage.

    Codes P000-P009 report on non-adherence to rule design principles.

    Specifically, the `PrincipleChecker` checks for the following:
        P000: No Limited Proxy, the rule does not detect a characteristic that relates directly to a malicious action,
        making it potentially noisy.

        P001: No Successful Malicious Action, the rule does not distinguish between successful and unsuccessful malicious
        actions, making it potentially noisy.

        P002: No Alert Throttling, the rule does not utilize the threshold limit option` to prevent alert flooding,
        making it potentially noisy.

        P003: No Exceptions, the rule does not include any exceptions for commom benign traffic,
        making it potentially noisy.

        P004: No Generalized Characteristic, the rule does detect a characteristic that is so specific
        that it is unlikely generalize.

        P005: No Generalized Position, the rule does detect the characteristic in a fixed position
        that and is unlikely to generalize as a result.
    """

    codes = (
        "P000",
        "P001",
        "P002",
        "P003",
        "P004",
        "P005",
    )

    def _check_rule(
        self: "PrincipleChecker",
        rule: idstools.rule.Rule,
    ) -> ISSUES_TYPE:
        issues: ISSUES_TYPE = []

        if count_rule_options(
            rule, ALL_DETECTION_KEYWORDS
        ) == 0 or is_rule_option_equal_to_regex(rule, "msg", _PROXY_MSG_REGEX):
            issues.append(
                Issue(
                    code="P000",
                    message=get_message("P000"),
                ),
            )

        if (
            self.__is_rule_initiated_internally(rule) is False
            and self.__does_rule_account_for_server_response(rule) is False
            and self.__does_rule_account_for_internal_content(rule) is False
            and self.__is_rule_stateful(rule) is False
        ):
            issues.append(
                Issue(
                    code="P001",
                    message=get_message("P001"),
                ),
            )

        if not self.__is_rule_threshold_limited(rule):
            issues.append(
                Issue(
                    code="P002",
                    message=get_message("P002"),
                ),
            )

        if not self.__does_rule_have_exceptions(rule):
            issues.append(
                Issue(
                    code="P003",
                    message=get_message("P003"),
                ),
            )

        if (
            count_rule_options(rule, "content") == 0
            and not count_rule_options(
                rule,
                set(SIZE_KEYWORDS)
                .union(CONTENT_KEYWORDS)
                .union(OTHER_PAYLOAD_KEYWORDS),
            )
            > 1
        ) or (
            is_rule_option_equal_to_regex(rule, "msg", _SPECIFIC_MSG_REGEX)
            and not is_rule_option_set(rule, "pcre")
        ):
            issues.append(
                Issue(
                    code="P004",
                    message=get_message("P004"),
                ),
            )

        if self.__has_fixed_http_uri_query_parameter_location(
            rule
        ) or self.__has_single_match_at_fixed_location(rule):
            issues.append(
                Issue(
                    code="P005",
                    message=get_message("P005"),
                ),
            )

        return issues

    @staticmethod
    def __is_rule_initiated_internally(
        rule: idstools.rule.Rule,
    ) -> Optional[bool]:
        if get_rule_option(rule, "proto") in ("ip",):
            return None

        dest_addr = get_rule_option(rule, "dest_addr")
        assert dest_addr is not None
        if (
            dest_addr not in ("any", "$EXTERNAL_NET")
            and IP_ADDRESS_REGEX.match(dest_addr) is None
        ):
            if is_rule_suboption_set(
                rule, "flow", "from_server"
            ) or is_rule_suboption_set(rule, "flow", "to_client"):
                return True

        source_addr = get_rule_option(rule, "source_addr")
        assert source_addr is not None
        if (
            source_addr not in ("any", "$EXTERNAL_NET")
            and IP_ADDRESS_REGEX.match(source_addr) is None
        ):
            if is_rule_suboption_set(
                rule, "flow", "to_server"
            ) or is_rule_suboption_set(rule, "flow", "from_client"):
                return True
            if is_rule_option_set(rule, "dns.query") or is_rule_option_set(
                rule,
                "dns_query",
            ):
                return True

        return False

    @staticmethod
    def __does_rule_account_for_server_response(
        rule: idstools.rule.Rule,
    ) -> Optional[bool]:
        if get_rule_option(rule, "proto") in ("ip",):
            return None

        if is_rule_suboption_set(rule, "flow", "from_server") or is_rule_suboption_set(
            rule, "flow", "to_client"
        ):
            return True

        msg = get_rule_option(rule, "msg")
        assert msg is not None
        if "response" in msg.lower():
            return True

        return False

    @staticmethod
    def __does_rule_account_for_internal_content(
        rule: idstools.rule.Rule,
    ) -> bool:
        source_addr = get_rule_option(rule, "source_addr")
        assert source_addr is not None
        if (
            source_addr not in ("any", "$EXTERNAL_NET")
            and IP_ADDRESS_REGEX.match(source_addr) is None
        ):
            return True

        return False

    @staticmethod
    def __is_rule_stateful(
        rule: idstools.rule.Rule,
    ) -> Optional[bool]:
        if (
            is_rule_option_equal_to_regex(rule, "flowbits", _BITS_ISSET_REGEX)
            or is_rule_option_equal_to_regex(rule, "flowint", _FLOWINT_ISSET_REGEX)
            or is_rule_option_equal_to_regex(rule, "xbits", _BITS_ISSET_REGEX)
        ):
            return True

        # flowbits.isnotset is used to reduce false positives as well, so it does not neccesarily indicate a stateful rule.
        if (
            is_rule_option_equal_to_regex(rule, "flowbits", _BITS_ISNOTSET_REGEX)
            or is_rule_option_equal_to_regex(rule, "flowint", _FLOWINT_ISNOTSET_REGEX)
            or is_rule_option_equal_to_regex(rule, "xbits", _BITS_ISNOTSET_REGEX)
        ):
            return True

        return False

    @staticmethod
    def __is_rule_threshold_limited(
        rule: idstools.rule.Rule,
    ) -> bool:
        value = get_rule_option(rule, "threshold")

        if value is None:
            return False

        if _THRESHOLD_LIMITED_REGEX.match(value) is not None:
            return True

        return False

    @staticmethod
    def __does_rule_have_exceptions(
        rule: idstools.rule.Rule,
    ) -> bool:
        positive_matches = 0
        negative_matches = 0

        for option_value in get_rule_options(rule, CONTENT_KEYWORDS):
            if option_value is None:
                continue
            if option_value.startswith("!"):
                negative_matches += 1
            else:
                positive_matches += 1

        if (
            positive_matches > 0 and negative_matches > 0
        ) or is_rule_option_equal_to_regex(
            rule,
            "flowbits",
            _FLOWBITS_ISNOTSET_REGEX,
        ):
            return True

        return False

    @staticmethod
    def __has_fixed_http_uri_query_parameter_location(
        rule: idstools.rule.Rule,
    ) -> bool:
        if count_rule_options(rule, "content") != 1:
            return False

        body = get_rule_body(rule)
        if _HTTP_URI_QUERY_PARAMETER_REGEX.match(body) is not None:
            return True

        return False

    @staticmethod
    def __has_single_match_at_fixed_location(
        rule: idstools.rule.Rule,
    ) -> bool:
        if (
            count_rule_options(rule, "content") == 2  # noqa: PLR2004
            and is_rule_option_set(rule, "http.method")
            and (
                is_rule_option_equal_to(rule, "content", "GET")
                or is_rule_option_equal_to(rule, "content", "POST")
            )
        ) and count_rule_options(rule, "content") != 1:
            return False

        contents = list(
            set(get_rule_options(rule, "content")).difference(['"GET"', '"POST"'])
        )
        if len(contents) != 1:
            return False
        content = contents[0]

        if is_rule_option_set(rule, "startswith"):
            return True

        if content is not None:
            # -2 to discard quotes
            length = len(content) - 2
        else:
            length = -1

        if (
            is_rule_option_equal_to(rule, "depth", str(length))
            or is_rule_option_equal_to(rule, "bsize", str(length))
            or (
                is_rule_option_set(rule, "http.uri")
                and is_rule_option_equal_to(rule, "urilen", str(length))
            )
        ):
            return True

        return False
